#########################################################################
#########################################################################
#########Example use PGEE fit with EN or STAT###########################
#########################################################################
#########################################################################

#load in PGEE functions
source("PGEE_solver_public.r")
library(mvtnorm)
library(matlab)

######
###Data generation (artificial example)
######

#SCENARIO3: rho1.1 = .5 rho1.2=0.5

tpm <- proc.time()
##########
###SIMULATIE DATA   
##########
beta_true <- c(1,1,-1,-1,0.5,0,0,0)
#structuur multicollineariteit
cov_mat1 <- diag(8)

r1.1 <- 0.5 
r1.1_mat <- matrix(c(1,r1.1,r1.1,1),2,2)

r1.2 <- 0.5
r1.2_mat <- matrix(c(1,r1.2,r1.2,1),2,2)

cov_mat1[1:2,1:2] <- r1.1_mat
cov_mat1[3:4,3:4] <- r1.2_mat


rho2 <- 0.7; #parameter longitudinale structuur 
indicator <- 1:5
distance <- abs(outer(indicator,indicator,"-"))
cov_mat2 <- rho2 ^distance  #AR[1] correlation structure

S <- 1 #aantal simulaties , we only take on example


set.seed(220687)
SIMDATA <- as.list(1:S)
for (s in 1:S)
{
#training set: 20 ind, 5 obs per individu: 100obs
X_tr <- rmvnorm(n=100,sigma = cov_mat1) #no time dependence of X-variables
Y_tr <- X_tr%*%beta_true
rand <- matrix(nrow=0,ncol=1)
while(length(rand) < 100)
{rand <- rbind(rand,matrix(rmvnorm(n=1,sigma= cov_mat2),nrow=5,ncol=1))} 
Y_tr <- Y_tr + rand  
id_tr <- rep(1:20,rep(5,20))
data_tr <- data.frame(id_tr,X_tr,Y_tr)
colnames(data_tr) <- c('id',paste('X_tr',1:8,sep=""),'Y_tr')


#in lijst opslaan
SIMDATA[[s]] <- data_tr
}

#############################################################################################################################################################

######
###Data preparation (to use in model: structure and standardization)
######

#not necessary in the exaple since we working on standardized data 
#in general reconmended to standardize both X and Y variables because penalization is not scale independent
# structure very important, first column should be an id variable which is numeric, second variables should be the X variables and the last column should be the response Y
#the algorithm works only with numerical variables, so if you have categories you have constract numerical indicators first before applying PGE estimator

######
###EN fit
######

#object needed, initialize
#data #dataframe met [id,X,Y]
method <- 'EN' #this elastic net penalty is used #'EN', 'BRIDGE' or 'SCAD'
alpha <- 0.9 #parameter 1 = lasso/SCAD, 0 =ridge controls the amount of penalty devided by sparceness and grouping effect producing penalties
lambda <- 20 #amount of penalization used
#V
M_it <- 500 # the maximum number of iterations , usually convergence is rather quick
tresh_beta <- 0.0001  # a numerical constant used to put parameters to zero if it fals below a certain figure, choose a small value #for stability if a betaj closer than this treshold, set to zero
M_diff <- 0.00001 # numaric constants to stop the algorithm is the norm of the difference in estimated beta vector is smaller than this constant(convergence indicator= precision)
k <- 0 #initialisation parameter for first fit, if numarical problems with initialization use a larger value (parameter van initile RIDGE/DECORRELATIE FIT, put to zero or a small value if it doesnt work increase)
option <- '' #'adaptive'  to get adaptive LASSO/EN  and anything else to get nonadaptive version which is recommended
distr <- "normal" #indicates the distribution used, options: "normal" or "binomial"
p=8 #the number of covariates 

EN_fit_result <-  PEE_LI2(data=data_tr,method='EN')

######
###SCAD fit
######

#object needed, initialize
#data #dataframe met [id,X,Y]
method <- 'SCAD' #this elastic net penalty is used #'EN', 'BRIDGE' or 'SCAD'
alpha <- 0.95 #parameter 1 = lasso/SCAD, 0 =ridge controls the amount of penalty devided by sparceness and grouping effect producing penalties
lambda <- 50 #amount of penalization used
a <- 3.7 #additional parameter, keep fixed as recommended buy Fan and Li
#V
M_it <- 500 # the maximum number of iterations , usually convergence is rather quick
tresh_beta <- 0.0001  # a numerical constant used to put parameters to zero if it fals below a certain figure, choose a small value #for stability if a betaj closer than this treshold, set to zero
M_diff <- 0.00001 # numaric constants to stop the algorithm is the norm of the difference in estimated beta vector is smaller than this constant(convergence indicator= precision)
k <- 0 #initialisation parameter for first fit, if numarical problems with initialization use a larger value (parameter van initile RIDGE/DECORRELATIE FIT, put to zero or a small value if it doesnt work increase)
option <- '' #'adaptive'  to get adaptive LASSO/EN  and anything else to get nonadaptive version which is recommended
distr <- "normal" #indicates the distribution used, options: "normal" or "binomial"
p=8 #the number of covariates 

SCAD_fit_result <-  PEE_LI2(data=data_tr,method='SCAD')

#########################################################################################################################################################

######
###Tuning parameter selection  by (leaf one subject out) cross-validation: example SCAD parameter
######

#generate grid of penalization parameters to evaluate
library(matlab)
n_lambda <- 20
n_alpha <- 5
lambda_vec <- linspace(0.0025,20,n=n_lambda)
alpha_vec <- rev(linspace (0.001,1,n_alpha)) #beginnen waarmee meest problemen verwacht
grid <- expand.grid(lambda_vec,alpha_vec)
names(grid) <- c('lambda' , 'alpha')
n_grid <- length(grid[,1])



#instellingen

M_it <- 1000 #(genoeg om convergentie te garanderen, indien convergentie stopt het sneller)
tresh_beta <- 0.00001 #for stability small enough, lots of coeff if a betaj closer than this treshold, set to zero (aanpassing: M_diff <<tresh_beta)
M_diff <- 0.0000000000000001 #(convergence indicator= precision)
method <- 'SCAD'
option <- 'nonadaptive'
p <- 8
a <- 3.7
k <- 1 #ridge initialisation

#objecten
MSE_CV <- array(dim=c(n_grid))
sde_CV <-  array(dim=c(n_grid))
MedianSE_CV <-  array(dim=c(n_grid))
n_subj <- 20  #aanpassen aan aantal landen
beta_cv <- array(dim=c(n_grid,n_subj,p))
conv_array <- array(dim=c(n_grid,n_subj))

for (i_grid in 1:n_grid)
  {
   lambda <- grid[i_grid,1]
   alpha <- grid[i_grid,2]

   resultaat <-  CV.PEE_LI(data=data_tr)
   MSE_CV[i_grid] <- resultaat$MSE_CV
   MedianSE_CV[i_grid] <- resultaat$MedianSE_CV
   sde_CV[i_grid] <- resultaat$sde
   beta_cv[i_grid,,] <- resultaat$beta_cv
   conv_array[i_grid,] <- resultaat$conv_cv
  }



###############################################################################################################################

######
###Penalization path: example:  EN path alpha = 0.5
######

#settings
library(matlab)
alpha <- 0.6
n_lambda <- 100
lambda_vec <- logspace(-2,5,n_lambda)  #vector of

#PGEE functional settings
a <- 3.7
M_it <- 1000
tresh_beta <- 0.00001
M_diff <- 0.000000000001
k <- 0  #OLS initialization
option <- 'nonadaptive' #non adaptive elastic net, no additional weights
method <- 'EN'  #method is elastic net as apposed to
p <- 8    #number of variables (all numeric)

#objecten
conv_mat <- vector(length=n_lambda)
Fit_EN <- array(dim=c(n_lambda,p))


#calculations
for(i_lambda in 1:n_lambda)
{
lambda <- lambda_vec[i_lambda]
fit <- PEE_LI2(data=data_tr,method)
Fit_EN[i_lambda,] <- as.vector(fit$`beta elastic net`)
conv_mat[i_lambda] <- fit$diff_beta
}

#Convergence
max(conv_mat) #no convergence problems

colnames(Fit_EN) <- 1:8#no variable names in artificial example

#selection of variables not always zero
csum <- colSums(Fit_EN)
Fit_EN_select <- Fit_EN[,csum != 0]

 #sort on initial size

 volg <- order(abs(Fit_EN_select[1,]))
 Fit_plot <- Fit_EN_select[,rev(volg)]


#plot tyhe 8 largest ones


linetype <- rep(c(1,2,6),100)
#beta1
plot(x=lambda_vec,y=Fit_plot[,1] ,
type='l',xlab=expression(lambda),ylab=expression(beta),
ylim=c(min(Fit_EN_select),max(Fit_EN_select)),lwd=2,lty=linetype[1],log='x', xlim = c(2,0.0085))

for(i in 2:8)
{
lines(x=lambda_vec,y=Fit_plot[,i],type='l',lwd=2,lty=linetype[i],log='x')
}